-module(ksudoku).
-author('hanak@inf.bme.hu, kapolnai@iit.bme.hu, modified by: Andras Eisenberger, Balazs Galambosi').
-vsn('$LastChangedDate: 2014-10-31 11:03:34 +0100 (Fri, 31 Oct 2014) $$').

% -compile(export_all).
-export([sudoku_be/1,sudoku_ki/2,megold/2,stopper/2,stopper/3, help/0,teljes_teszt/1,teszt/2]).

-import(sudoku,[sudoku/1]).

%% @type sspec() = {dist(), board()}.
%% @type dist()  = integer().
%% @type field() = [info()].
%% @type info()  = s | w | integer().
%% @type board() = [[field()]].

%% @type ssol() = [[integer()]].

%% @spec sudoku:sudoku(SSpec::sspec()) -> SSols::[ssol()].
%% @doc  SSols az SSpec feladvnyt kielgt megoldsok listja.

%% ksudoku-specifikcik

%% @spec sudoku_be(FileIn::string()) -> SSpec::sspec().
%% @doc  SSpec a FileIn szvegfjlbl beolvasott Sudoku-feladvny.
%%
sudoku_be(FileIn) ->
    case file:open(FileIn,read) of
	{ok, InputFile} ->
	    {N,_} = case io:get_line(InputFile,'') of
			eof ->
			    io:format("Ures a(z) ~s fajl.~n",[FileIn]),
			    exit('Hibas feladvany');
			I ->
			    string:to_integer(cleaned(I))
		    end,
	    T = tabla_be(InputFile),
	    file:close(InputFile),
%	    io:fwrite("~w\n",[T]),
	    KK = length(T),
	    if N < 1 orelse KK < N ->
		    io:format("a tavolsagkonstans kisebb 1-nel vagy nagyobb a sorhossznal.~n",[]),
		    exit('Hibas feladvany');
	       true ->
		    case helyes_feladvany(KK,T) of
			true ->
			    {N,T};
			false ->
			    io:format("Hibas a feladvany a(z) ~s fajlban.~n",[FileIn]),
			    exit('Hibas feladvany')
		    end
	    end;
	_ ->
	    io:format("Nincs ~s nevu fajl.~n",[FileIn]),
	    exit('Nincs ilyen feladvany')
    end.

tabla_be(InputFile) ->
    tabla_be(InputFile,[]).

tabla_be(InputFile,Zss) ->
    case io:get_line(InputFile,'') of
	eof ->
	    KK = length(Zss),
	    if KK < 1 orelse 25 < KK ->
		    io:format("1-nel kisebb vagy 5-nel nagyobb a cellameret.~n",[]),
		    exit('Hibas feladvany');
	       true ->
		    case lists:all(fun(L) -> length(L) == KK end, Zss) of
			false ->
			    io:format("Nem negyzetes a feladvany, vagy nem egyforma hosszuak a sorok.~n",[]),
			    exit('Hibas feladvany');
			true ->
			    lists:reverse(Zss)
		    end
	    end;
	Sor ->
	    Ss = string:tokens(cleaned(Sor)," "),
	    Zs = lists:map(fun str2infolist/1, Ss),
	    case Zs of
		[] ->
		    tabla_be(InputFile,Zss);
		_ ->
		    tabla_be(InputFile,[Zs|Zss])
	    end
    end.

cleaned(Sor) ->
    S0 = lists:filter(fun(C) -> C /= $\t end,Sor),
    string:strip(string:strip(S0),right,$\n).

str2infolist([$-]) ->
    [];
str2infolist(XXs=[X|_Xs]) when $0=<X andalso X=<$9 ->
    {I,Rs} = string:to_integer(XXs),
    [I|str2infolist(Rs)];
str2infolist([C|Xs]) ->
    [list_to_atom([C])|str2infolist(Xs)];
str2infolist([]) ->
    [].

%% @spec helyes_feladvany(Size::integer(),Board::board()) -> B::bool().
%% @doc  B igaz, ha a feladvany megfelel a specifikacionak.
%%
helyes_feladvany(KK,T) ->
    lists:all(fun(Fs) -> helyes_mezo(KK,Fs) end, lists:concat(T)).

helyes_mezo(KK,Fs) ->
    helyes_mezo(KK,Fs,0,0,0).

helyes_mezo(_KK,[],S,V,W) ->
    S =< 1 andalso V =< 1 andalso W =< 1;
helyes_mezo(KK,[s|Fs],S,V,W) ->
    helyes_mezo(KK,Fs,S+1,V,W);
helyes_mezo(KK,[w|Fs],S,V,W) ->
    helyes_mezo(KK,Fs,S,V,W+1);
helyes_mezo(KK,[I|Fs],S,V,W) when is_integer(I) ->
    1 =< I andalso I =< KK andalso helyes_mezo(KK,Fs,S,V+1,W);
helyes_mezo(_KK,[_|_Fs],_S,_V,_W) ->
    io:format("Hibas info van az input fajlban.~n",[]),
    exit('Hibas feladvany').

tablak_ki(OutputFile,[SSol]) ->
    io:fwrite(OutputFile,"~s~n",[tabla_ki(SSol)]);
tablak_ki(OutputFile,[SSol|SSols]) ->
    io:fwrite(OutputFile,"~s~n~n",[tabla_ki(SSol)]),
    tablak_ki(OutputFile,SSols);
tablak_ki(OutputFile,[]) ->
    io:fwrite(OutputFile,"",[]).

tabla_ki([L]) ->
    sor_ki(L);
tabla_ki([L|Ls]) ->
    sor_ki(L) ++ "\n" ++ tabla_ki(Ls);
tabla_ki(_) ->
    [].

sor_ki([H|T]) ->
    I2L = fun(Int) -> io_lib:format(if length([H|T]) > 9 -> "~2w" ; true -> "~w" end, [Int]) end,
    I2L(H) ++ lists:foldr(fun(A,B) -> " " ++ I2L(A) ++ B end,"",T);
sor_ki([]) ->
    "".

%% @spec sudoku_ki(Mode::mode(),FileOut::string(),SSols::[ssol()])
%%                                                               -> void().
%% @doc  Az SSols listaban atadott Sudoku-megoldasokat a Mode szerinti
%%       alakban kiirja a FileOut szovegfajlba.
%%
sudoku_ki(Mode,FileOut,SSols) ->
    {ok,OutputFile} = file:open(FileOut,write),
    case Mode of
	plstr ->
	    io:fwrite(OutputFile,"~w.~n",[SSols]);
	splstr ->
	    io:fwrite(OutputFile,"~w.~n",[lists:sort(SSols)]);
	spretty ->
	    tablak_ki(OutputFile,lists:sort(SSols))
    end,
    file:close(OutputFile).

sudoku_ki(FileOut,SSols) ->
    sudoku_ki(spretty,FileOut,SSols).


%% @spec sudokusol_be(FileIn::string()) -> SSols::[ssol()].
%% @doc  SSpec a FileIn szvegfjlbl beolvasott Sudoku-feladvny.
%%       Ksztette: Eisenberger Andrs.
sudokusol_be(FileSol) ->
    case file:open(FileSol,read) of
        {ok, InputFile}      -> ok;
        {error, InputFile=R} -> throw({error,R,file:format_error(R)})
    end,
    case io:read(InputFile, "") of
        {ok, Sol}       -> ok;
        {error, Sol=R2} -> throw({error,R2,file:format_error(R2)})
    end,
    file:close(InputFile),
    Sol.



%% @spec ksudoku:megold(FileIn::string(),FileOut::string()) -> void()
%% @doc  Beolvas egy feladvnyt a FileIn szvegfjlbl s sszes
%%       megoldst kirja a FileOut szvegfjlba. Ehhez
%%       felhasznlja a sudoku:sudoku/1 fggvnyt.
%%
megold(FileIn,FileOut) ->
    KT = sudoku_be(FileIn),
    SSols = sudoku:sudoku(KT), 
    sudoku_ki(spretty,FileOut,SSols).

%% @spec ksudoku:stopper(FileIn::string(),FileOut::string()) -> void()
%% @doc  Mint ksudoku:megold/2, de a vgn kirja a FileIn nevet,
%%       a megoldsok szmt s a futsi idt is.
stopper(FileIn,FileOut) -> stopper(FileIn,FileOut,"").

%% @spec ksudoku:stopper(FileIn::string(),FileOut::string(),FileSol::string()) -> void()
%% @doc  Mint ksudoku:megold/2, de a vgn kirja a FileIn nevet,
%%       a megoldsok szmt s a futsi idt is. Ha lehet, sszeveti FileSol megoldssal is.
stopper(FileIn,FileOut,FileSol) ->
    KT = sudoku_be(FileIn),
    erlang:garbage_collect(),timer:sleep(10),
    _T1 = statistics(runtime),
    SSols = sudoku:sudoku(KT),
    {_,T2} = statistics(runtime),
    sudoku_ki(FileOut,SSols),
    io:format("Feladvany: ~s, ", [FileIn]),
    try   % Eisenberger Andrs kiegsztse
        case lists:sort(sudokusol_be(FileSol)) =:= lists:sort(SSols) of
            true  -> io:format("HELYES ");
            false -> io:format("ROSSZ ")
        end
    catch
        {error, _, _} -> error%io:format("NINCS MO ")
    end,
    io:format("megoldas: ~3w, futasi ido: ",
	      [length(SSols)]),
    if T2 >= 10000 -> io:format("~5w s\n", [ T2 div 1000 ]);
%       T2 >= 1000 -> io:format("~5.2f s\n", [ T2 / 1000 ]);
       true        -> io:format("~5w ms\n", [ T2 ])
    end.


%% ------- experimental -----


help() ->
    io:format(
      "teljes_teszt(Timeout): meg kiserleti stadiumban van."
      "Megoldja az osszes, tests/testXXXd.txt specifikaciot idolimittel.~n"
      "Pelda: futtatas bash-bol, 10s idolimit, megoldasok tests/testXXXt.txt fajlokba: ~n"
      "$ erl -noshell -eval 'ksudoku:teljes_teszt(10).' -s init stop~n"
     ).

%% @spec teljes_teszt(Timeout::integer()) -> [done|timeout].
%% @doc A 'tests' knyvtrban lev sszes "testXXXd.txt" tesztllomny esetn
%%  - lefuttatja a tesztet Timeout msodperces idkorlttal,
%%  - ellenrzi, hogy a testXXXs.txt llomnyban megadott megoldshalmazt kapta,
%%  - olvashat formban (lsd megold/2) kirja az eredmnyt a 'tests_out_er'
%%    knyvtr testXXXt.txt nev llomnyba.
%% Az llomnynevekben az XXX szm tetszleges hosszsg lehet.
teljes_teszt(Timeout) ->
    Dir = "tests/",
    ODir = "tests_out_er/",
    case file:make_dir(ODir) of
        ok              -> ok;
        {error, eexist} -> ok;
        {error, Rd}     -> throw({error,Rd,file:format_error(Rd)})
    end,
    case file:list_dir(Dir) of
        {error, R} ->
            throw({error,R,file:format_error(R)});
        {ok, Content} ->
            IOFiles = [{Dir++Input,ODir++Output,Dir++Solution}
                       || Input <- lists:sort(Content),
                          {match,_} <- [re:run(Input, "test[0-9]*d\.txt")],
                          Solution <- [re:replace(Input, "d\.txt\$", "s.txt", [{return, list}])],
                          Output <- [re:replace(Input, "d\.txt\$", "t.txt", [{return, list}])]
                      ],
            teszt(Timeout, IOFiles)
    end.

%% @spec teszt(Timeout::integer(),
%%            IOFiles::[{Input::string(),Output::string(),Solution::string()}]) -> [done|timeout].
%% @doc Megoldja az sszes Input fjlban tallhat specifikcit Timeout idlimittel,
%%      megoldsokat az Output fjlokba teszi. Ha lehet, sszeveti a megoldsokat Solution-nel.
teszt(Timeout, IOFiles) ->
    [begin
         process_flag(trap_exit, true),
         Worker = spawn_link(fun() -> stopper(Input, Output, Solution) end),
         Timer  = spawn_link(fun() -> timer:sleep(Timeout * 1000) end),
         receive
             {'EXIT', Worker, _Reason} -> 
                 exit(Timer, kill),
                 receive {'EXIT', Timer, killed} -> ok end, % feldolgozzuk a halalat
                 done;
             {'EXIT', Timer, _Reason} ->
                 exit(Worker, kill),
                 io:format("Feladvany: ~s -- Timeout (~p s)~n", [Input,Timeout]),
                 receive {'EXIT', Worker, killed} -> ok end, % feldolgozzuk a halalat
                 timeout
         end
     end
     || {Input, Output, Solution} <- IOFiles
    ].
